import random
from collections import deque
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Iterator

import numpy as np
from mcap.reader import make_reader
from mcap_ros2.decoder import DecoderFactory
from pytest_benchmark.fixture import BenchmarkFixture
from rosbags.highlevel import AnyReader
from rosbags.rosbag2 import StoragePlugin, Writer
from rosbags.typesys import Stores, get_typestore

from pybag.mcap_reader import McapFileReader


def create_test_mcap(path: Path, message_count: int = 1000, seed: int = 0) -> Path:
    """Create an MCAP file with a single `/odom` topic of Odometry messages."""
    rng = random.Random(seed)

    typestore = get_typestore(Stores.LATEST)
    Odometry = typestore.types["nav_msgs/msg/Odometry"]
    Header = typestore.types["std_msgs/msg/Header"]
    Time = typestore.types["builtin_interfaces/msg/Time"]
    PoseWithCovariance = typestore.types["geometry_msgs/msg/PoseWithCovariance"]
    Pose = typestore.types["geometry_msgs/msg/Pose"]
    Point = typestore.types["geometry_msgs/msg/Point"]
    Quaternion = typestore.types["geometry_msgs/msg/Quaternion"]
    TwistWithCovariance = typestore.types["geometry_msgs/msg/TwistWithCovariance"]
    Twist = typestore.types["geometry_msgs/msg/Twist"]
    Vector3 = typestore.types["geometry_msgs/msg/Vector3"]

    with Writer(path, version=9, storage_plugin=StoragePlugin.MCAP) as writer:
        odom_conn = writer.add_connection("/odom", Odometry.__msgtype__, typestore=typestore)
        for i in range(message_count):
            timestamp = int(i * 1_500_000_000)
            odom = Odometry(
                header=Header(
                    stamp=Time(
                        sec=timestamp // 1_000_000_000,
                        nanosec=timestamp % 1_000_000_000,
                    ),
                    frame_id="map"
                ),
                child_frame_id="base_link",
                pose=PoseWithCovariance(
                    pose=Pose(
                        position=Point(
                            x=rng.random(),
                            y=rng.random(),
                            z=rng.random()
                        ),
                        orientation=Quaternion(
                            x=rng.random(),
                            y=rng.random(),
                            z=rng.random(),
                            w=rng.random(),
                        ),
                    ),
                    covariance=np.array([rng.random() for _ in range(36)]),
                ),
                twist=TwistWithCovariance(
                    twist=Twist(
                        linear=Vector3(
                            x=rng.random(),
                            y=rng.random(),
                            z=rng.random(),
                        ),
                        angular=Vector3(
                            x=rng.random(),
                            y=rng.random(),
                            z=rng.random(),
                        ),
                    ),
                    covariance=np.array([rng.random() for _ in range(36)]),
                ),
            )
            writer.write(odom_conn, timestamp, typestore.serialize_cdr(odom, Odometry.__msgtype__))

    return next(Path(path).rglob("*.mcap"))


def read_with_pybag(mcap: Path) -> Iterator[Any]:
    reader = McapFileReader.from_file(mcap)
    for topic in reader.get_topics():
        for message in reader.messages(topic):
            yield message.data


def read_with_rosbags(mcap: Path) -> Iterator[Any]:
    typestore = get_typestore(Stores.LATEST)
    with AnyReader([mcap.parent]) as reader:
        for conn, _, data in reader.messages():
            yield typestore.deserialize_cdr(data, conn.msgtype)


def read_with_official(mcap: Path) -> Iterator[Any]:
    with open(mcap, "rb") as f:
        reader = make_reader(f, decoder_factories=[DecoderFactory()])
        for _, _, _, ros_msg in reader.iter_decoded_messages(log_time_order=False):
            yield ros_msg


def test_official_read(benchmark: BenchmarkFixture) -> None:
    with TemporaryDirectory() as tmpdir:
        mcap = create_test_mcap(Path(tmpdir) / "test")
        benchmark(lambda: deque(read_with_official(mcap), maxlen=0))


def test_rosbags_read(benchmark: BenchmarkFixture) -> None:
    with TemporaryDirectory() as tmpdir:
        mcap = create_test_mcap(Path(tmpdir) / "test")
        benchmark(lambda: deque(read_with_rosbags(mcap), maxlen=0))


def test_pybag_read(benchmark: BenchmarkFixture) -> None:
    with TemporaryDirectory() as tmpdir:
        mcap = create_test_mcap(Path(tmpdir) / "test")
        benchmark(lambda: deque(read_with_pybag(mcap), maxlen=0))



"""Test filtering of messages using a lambda."""
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from rosbags.rosbag2 import StoragePlugin, Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.store import Typestore

from pybag.mcap_reader import McapFileReader


def _find_mcap_file(temp_dir: str) -> Path:
    return next(Path(temp_dir).rglob("*.mcap"))


@pytest.fixture(params=[Stores.ROS2_JAZZY, Stores.ROS2_HUMBLE])
def typestore(request):
    return get_typestore(request.param)


def test_messages_filter(typestore: Typestore):
    # Write a temporary mcap file
    Int32 = typestore.types["std_msgs/msg/Int32"]
    with TemporaryDirectory() as temp_dir:
        with Writer(
            Path(temp_dir) / "rosbags",
            version=9,
            storage_plugin=StoragePlugin.MCAP,
        ) as writer:
            conn = writer.add_connection("/rosbags", Int32.__msgtype__, typestore=typestore)
            writer.write(conn, 0, typestore.serialize_cdr(Int32(data=1), Int32.__msgtype__))
            writer.write(conn, 1, typestore.serialize_cdr(Int32(data=-1), Int32.__msgtype__))

        mcap_file = _find_mcap_file(temp_dir)
        reader = McapFileReader.from_file(mcap_file)

        all_messages = list(reader.messages("/rosbags"))
        assert len(all_messages) == 2
        assert all_messages[0].data.data == 1
        assert all_messages[1].data.data == -1

        positive = list(reader.messages("/rosbags", filter=lambda m: m.data.data > 0))
        assert len(positive) == 1
        assert positive[0].data.data == 1

        negative = list(reader.messages("/rosbags", filter=lambda m: m.data.data < 0))
        assert len(negative) == 1
        assert negative[0].data.data == -1


import logging
from abc import ABC, abstractmethod
from collections import namedtuple
from pathlib import Path
from typing import Generator

from pybag.crc import assert_crc
from pybag.io.raw_reader import BaseReader, BytesReader, FileReader
from pybag.mcap.error import (
    McapNoStatisticsError,
    McapNoSummaryIndexError,
    McapNoSummarySectionError,
    McapUnknownCompressionError,
    McapUnknownSchemaError
)
from pybag.mcap.record_parser import (
    FOOTER_SIZE,
    MAGIC_BYTES_SIZE,
    McapRecordParser,
    McapRecordType
)
from pybag.mcap.records import (
    ChannelRecord,
    ChunkIndexRecord,
    ChunkRecord,
    FooterRecord,
    HeaderRecord,
    MessageIndexRecord,
    MessageRecord,
    SchemaRecord,
    StatisticsRecord
)

logger = logging.getLogger(__name__)


def decompress_chunk(chunk: ChunkRecord, *, check_crc: bool = False) -> bytes:
    """Decompress the records field of a chunk."""
    if chunk.compression == 'zstd':
        import zstandard as zstd
        chunk_data = zstd.ZstdDecompressor().decompress(chunk.records)
    elif chunk.compression == 'lz4':
        import lz4.frame
        chunk_data = lz4.frame.decompress(chunk.records)
    elif chunk.compression == '':
        chunk_data = chunk.records
    else:
        error_msg = f'Unknown compression type: {chunk.compression}'
        raise McapUnknownCompressionError(error_msg)

    # Validate the CRC if requested
    if check_crc and chunk.uncompressed_crc != 0:
        assert_crc(chunk_data, chunk.uncompressed_crc)
    return chunk_data


# TODO: Is this the minimal set of methods needed?
class BaseMcapRecordReader(ABC):
    @abstractmethod
    def close(self) -> None:
        """Close the MCAP file and release all resources."""
        ...

    @abstractmethod
    def get_header(self) -> HeaderRecord:
        """Get the header record from the MCAP file."""
        ...

    @abstractmethod
    def get_footer(self) -> FooterRecord:
        """Get the footer record from the MCAP file."""
        ...

    @abstractmethod
    def get_statistics(self) -> StatisticsRecord:
        """Get the statistics record from the MCAP file."""
        ...

    # Schema Management

    @abstractmethod
    def get_schemas(self) -> dict[int, SchemaRecord]:
        """Get all schemas defined in the MCAP file."""
        ...

    @abstractmethod
    def get_schema(self, schema_id: int) -> SchemaRecord | None:
        """Get a schema by its ID."""
        ...

    @abstractmethod
    def get_channel_schema(self, channel_id: int) -> SchemaRecord | None:
        """Get the schema for a given channel ID."""
        ...

    @abstractmethod
    def get_message_schema(self, message: MessageRecord) -> SchemaRecord:
        """Get the schema for a given message."""
        ...

    # Channel Management

    @abstractmethod
    def get_channels(self) -> dict[int, ChannelRecord]:
        """Get all channels/topics in the MCAP file."""
        ...

    @abstractmethod
    def get_channel(self, channel_id: int) -> ChannelRecord | None:
        """Get a channel by its ID."""
        ...

    @abstractmethod
    def get_channel_id(self, topic: str) -> int | None:
        """Get a channel ID by its topic."""
        ...

    # Message Index Management

    @abstractmethod
    def get_message_indexes(self, chunk_index: ChunkIndexRecord) -> dict[int, MessageIndexRecord]:
        """Get all message indexes from the MCAP file."""
        ...

    @abstractmethod
    def get_message_index(self, chunk_index: ChunkIndexRecord, channel_id: int) -> MessageIndexRecord | None:
        """Get a message index for a given channel ID."""
        ...

    # Chunk Management

    @abstractmethod
    def get_chunk_indexes(self, channel_id: int | None = None) -> list[ChunkIndexRecord]:
        """Get all chunk indexes from the MCAP file."""
        ...

    @abstractmethod
    def get_chunk(self, chunk_index: ChunkIndexRecord) -> ChunkRecord:
        """Get a chunk by its index."""
        ...

    # Message Management

    @abstractmethod
    def get_message(
        self,
        channel_id: int,
        timestamp: int | None = None,
    ) -> MessageRecord | None:
        ...

    @abstractmethod
    def get_messages(
        self,
        channel_id: int,
        start_timestamp: int | None = None,
        end_timestamp: int | None = None,
    ) -> Generator[MessageRecord, None, None]:
        ...


class McapRecordRandomAccessReader(BaseMcapRecordReader):
    """Class to efficiently get records from an MCAP file.

    Args:
        file: The file to read from.
    """

    def __init__(self, file: BaseReader, *, check_crc: bool = False):
        self._file = file
        self._check_crc = check_crc

        self._version = McapRecordParser.parse_magic_bytes(self._file)
        logger.debug(f'MCAP version: {self._version}')

        footer = self.get_footer()

        # Summary section start
        self._summary_start = footer.summary_start
        logger.debug(f'Summary start: {self._summary_start}')
        if self._summary_start == 0:
            error_msg = 'No summary section detected in MCAP'
            raise McapNoSummarySectionError(error_msg)

        # Summary offset section start
        self._summary_offset_start = footer.summary_offset_start
        logger.debug(f'Summary offset start: {self._summary_offset_start}')
        if self._summary_offset_start == 0:
            error_msg = 'No summary offset section detected in MCAP'
            raise McapNoSummaryIndexError(error_msg)

        # Load summary offsets
        Offset = namedtuple('Offset', ['group_start', 'group_length'])
        self._summary_offset: dict[int, Offset] = {}
        self._file.seek_from_start(self._summary_offset_start)
        while McapRecordParser.peek_record(self._file) == McapRecordType.SUMMARY_OFFSET:
            record = McapRecordParser.parse_summary_offset(self._file)
            self._summary_offset[record.group_opcode] = Offset(record.group_start, record.group_length)

        # Caches for chunk and message indexes
        self._chunk_indexes: list[ChunkIndexRecord] | None = None
        self._message_indexes: dict[int, dict[int, MessageIndexRecord]] = {}

        # Cached schema and channel dictionaries populated on first access
        self._schemas: dict[int, SchemaRecord] | None = None
        self._channels: dict[int, ChannelRecord] | None = None

    # Helpful Constructors

    @staticmethod
    def from_file(file_path: Path | str) -> 'McapRecordRandomAccessReader':
        """
        Create a new MCAP reader from a file.
        """
        return McapRecordRandomAccessReader(FileReader(file_path))

    @staticmethod
    def from_bytes(data: bytes) -> 'McapRecordRandomAccessReader':
        """
        Create a new MCAP reader from a bytes object.
        """
        return McapRecordRandomAccessReader(BytesReader(data))

    # Destructors

    def close(self) -> None:
        """Close the MCAP file and release all resources."""
        self._file.close()

    # Context Managers

    def __enter__(self) -> 'McapRecordRandomAccessReader':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    # Getters for records

    def get_header(self) -> HeaderRecord:
        """Get the header record from the MCAP file."""
        self._file.seek_from_start(MAGIC_BYTES_SIZE)
        return McapRecordParser.parse_header(self._file)

    def get_footer(self) -> FooterRecord:
        """Get the footer record from the MCAP file."""
        self._file.seek_from_end(FOOTER_SIZE + MAGIC_BYTES_SIZE)
        return McapRecordParser.parse_footer(self._file)

    def get_statistics(self) -> StatisticsRecord:
        """Get the statistics record from the MCAP file."""
        if McapRecordType.STATISTICS not in self._summary_offset:
            raise McapNoStatisticsError('No statistics section detected in MCAP')
        self._file.seek_from_start(self._summary_offset[McapRecordType.STATISTICS].group_start)
        return McapRecordParser.parse_statistics(self._file)

    # Schema Management

    def get_schemas(self) -> dict[int, SchemaRecord]:
        """
        Get all schemas defined in the MCAP file.

        Returns:
            A dictionary mapping schema IDs to SchemaInfo objects.
        """
        if self._schemas is None:
            self._file.seek_from_start(self._summary_offset[McapRecordType.SCHEMA].group_start)
            schemas: dict[int, SchemaRecord] = {}
            while McapRecordParser.peek_record(self._file) == McapRecordType.SCHEMA:
                schema = McapRecordParser.parse_schema(self._file)
                if schema is None:  # Invalid schema, should be ignored
                    continue
                schemas[schema.id] = schema
            self._schemas = schemas
        return self._schemas

    def get_schema(self, schema_id: int) -> SchemaRecord | None:
        """
        Get a schema by its ID.

        Args:
            schema_id: The ID of the schema.

        Returns:
            The schema or None if the schema does not exist.
        """
        return self.get_schemas().get(schema_id)

    def get_channel_schema(self, channel_id: int) -> SchemaRecord | None:
        """
        Get the schema for a given channel ID.

        Args:
            channel_id: The ID of the channel.

        Returns:
            The schema of the channel or None if the channel/schema does not exist.
        """
        channel = self.get_channel(channel_id)
        if channel is None:
            return None
        return self.get_schema(channel.schema_id)

    def get_message_schema(self, message: MessageRecord) -> SchemaRecord:
        """
        Get the schema for a given message.

        Args:
            message: The message to get the schema for.

        Returns:
            The schema for the message.
        """
        schema = self.get_channel_schema(message.channel_id)
        if schema is None:
            raise McapUnknownSchemaError(f'Unknown schema for channel {message.channel_id}')
        return schema

    # Channel Management

    def get_channels(self) -> dict[int, ChannelRecord]:
        """
        Get all channels/topics in the MCAP file.

        Returns:
            A dictionary mapping channel IDs to channel information.
        """
        if self._channels is None:
            self._file.seek_from_start(self._summary_offset[McapRecordType.CHANNEL].group_start)
            channels: dict[int, ChannelRecord] = {}
            while McapRecordParser.peek_record(self._file) == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                channels[channel.id] = channel
            self._channels = channels
        return self._channels

    def get_channel(self, channel_id: int) -> ChannelRecord | None:
        """
        Get channel information by its ID.

        Args:
            channel_id: The ID of the channel.

        Returns:
            The channel information or None if the channel does not exist.
        """
        return self.get_channels().get(channel_id)

    def get_channel_id(self, topic: str) -> int | None:
        """Get a channel ID by its topic."""
        for channel in self.get_channels().values():
            if channel.topic == topic:
                return channel.id
        return None

    # Message Index Management

    def get_message_indexes(self, chunk_index: ChunkIndexRecord) -> dict[int, MessageIndexRecord]:
        """
        Get all message indexes from the MCAP file.

        Args:
            chunk_index: The chunk index to get the message indexes from.

        Returns:
            A list of MessageIndexRecord objects.
        """
        key = chunk_index.chunk_start_offset
        if key in self._message_indexes:
            return self._message_indexes[key]

        message_index: dict[int, MessageIndexRecord] = {}
        for channel_id, message_index_offset in chunk_index.message_index_offsets.items():
            self._file.seek_from_start(message_index_offset)
            message_index[channel_id] = McapRecordParser.parse_message_index(self._file)
            message_index[channel_id].records.sort(key=lambda x: x[0])

        self._message_indexes[key] = message_index
        return message_index

    def get_message_index(self, chunk_index: ChunkIndexRecord, channel_id: int) -> MessageIndexRecord | None:
        """
        Get a message index for a given channel ID.

        Args:
            chunk_index: The chunk index to get the message indexes from.
            channel_id: The ID of the channel.

        Returns:
            A MessageIndexRecord object or None if the channel does not exist for the chunk.
        """
        return self.get_message_indexes(chunk_index).get(channel_id)

    # Chunk Management

    def _load_chunk_indexes(self) -> None:
        if self._chunk_indexes is not None:
            return
        self._file.seek_from_start(self._summary_offset[McapRecordType.CHUNK_INDEX].group_start)
        self._chunk_indexes = []
        while McapRecordParser.peek_record(self._file) == McapRecordType.CHUNK_INDEX:
            chunk_index = McapRecordParser.parse_chunk_index(self._file)
            self._chunk_indexes.append(chunk_index)
        self._chunk_indexes.sort(key=lambda x: x.message_start_time)

    def get_chunk_indexes(self, channel_id: int | None = None) -> list[ChunkIndexRecord]:
        """
        Get all chunk indexes from the MCAP file.

        Args:
            channel_id: The ID of the channel to get the chunk indexes for.
                        If None, all chunk indexes are returned.

        Returns:
            A list of ChunkIndexRecord objects.
        """
        self._load_chunk_indexes()
        if self._chunk_indexes is None:
            return []
        if channel_id is None:
            return self._chunk_indexes
        return [ci for ci in self._chunk_indexes if channel_id in ci.message_index_offsets]

    def get_chunk(self, chunk_index: ChunkIndexRecord) -> ChunkRecord:
        """
        Get a chunk by its index.

        Args:
            chunk_index: The chunk index to get the chunk from.

        Returns:
            A ChunkRecord object.
        """
        self._file.seek_from_start(chunk_index.chunk_start_offset)
        return McapRecordParser.parse_chunk(self._file)

    # Message Management

    def get_message(
        self,
        channel_id: int,
        timestamp: int | None = None,
    ) -> MessageRecord | None:
        """
        Get a message from a given channel at a given timestamp.

        If the timestamp is not provided, the first message in the channel is returned.

        Args:
            channel_id: The ID of the channel.
            timestamp: The timestamp of the message.

        Returns:
            A MessageRecord object or None if the message does not exist.
        """
        chunk_indexes = self.get_chunk_indexes(channel_id)
        for chunk_index in chunk_indexes:
            if timestamp is None or (chunk_index.message_start_time <= timestamp <= chunk_index.message_end_time):
                # The message must be in the chunk based on the start and end times
                message_index = self.get_message_index(chunk_index, channel_id)
                if message_index is None:
                    return None

                # Make sure the timestamp is in the message index
                offset = next((r[1] for r in message_index.records if r[0] == timestamp), None)
                if offset is None:
                    return None

                # Read data from chunk
                chunk = self.get_chunk(chunk_index)
                reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
                reader.seek_from_start(offset)
                return McapRecordParser.parse_message(reader)
        return None

    def get_messages(
        self,
        channel_id: int,
        start_timestamp: int | None = None,
        end_timestamp: int | None = None,
    ) -> Generator[MessageRecord, None, None]:
        """
        Get all messages from a given channel.

        If the start and end timestamps are not provided, all messages in the channel are returned.

        Args:
            channel_id: The ID of the channel.
            start_timestamp: The start timestamp to filter by. If None, no filtering is done.
            end_timestamp: The end timestamp to filter by. If None, no filtering is done.

        Returns:
            A generator of MessageRecord objects.
        """
        for chunk_index in self.get_chunk_indexes(channel_id):
            # Skip chunk that do not match the timestamp range
            if start_timestamp is not None and chunk_index.message_end_time < start_timestamp:
                continue
            if end_timestamp is not None and chunk_index.message_start_time > end_timestamp:
                continue

            # Get the message index for the chunk
            message_index = self.get_message_index(chunk_index, channel_id)
            if message_index is None:
                continue

            # Read all messages in the chunk
            chunk = self.get_chunk(chunk_index)
            reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
            for timestamp, offset in message_index.records:
                # Skip messages that do not match the timestamp range
                if start_timestamp is not None and timestamp < start_timestamp:
                    continue
                if end_timestamp is not None and timestamp > end_timestamp:
                    continue

                # Read the message
                reader.seek_from_start(offset)
                yield McapRecordParser.parse_message(reader)

    # TODO: Low Priority
    # - Metadata Index
    # - Attachment Index


class McapRecordReaderFactory:
    """Factory to create a McapFileSequentialReader or McapFileRandomAccessReader."""

    @staticmethod
    def from_file(file_path: Path | str) -> McapRecordRandomAccessReader:
        """Create a new MCAP reader from a file."""
        try:
            # Try to create a random access reader first
            return McapRecordRandomAccessReader.from_file(file_path)
        except McapNoSummarySectionError:
            # If no summary section exists, fall back to sequential reader
            # TODO: Implement the sequential reader
            logger.warning('No summary section exists in MCAP, falling back to sequential reader')
            raise NotImplementedError('Sequential readers are not implemented yet')
        except McapNoSummaryIndexError:
            # If no summary index exists, fall back to sequential reader
            # TODO: Use sequential reader? Or just create the index?
            logger.warning('No summary index exists in MCAP, falling back to sequential reader')
            raise NotImplementedError('Sequential readers are not implemented yet')

    @staticmethod
    def from_bytes(data: bytes) -> McapRecordRandomAccessReader:
        """Create a new MCAP reader from a bytes object."""
        try:
            # Try to create a random access reader first
            return McapRecordRandomAccessReader.from_bytes(data)
        except McapNoSummarySectionError:
            # If no summary section exists, fall back to sequential reader
            # TODO: Implement the sequential reader
            logger.warning('No summary section exists in MCAP, falling back to sequential reader')
            raise NotImplementedError('Sequential readers are not implemented yet')
        except McapNoSummaryIndexError:
            # If no summary index exists, fall back to sequential reader
            # TODO: Use sequential reader? Or just create the index?
            logger.warning('No summary index exists in MCAP, falling back to sequential reader')
            raise NotImplementedError('Sequential readers are not implemented yet')



from dataclasses import is_dataclass
from typing import Any

from pybag.encoding import MessageEncoder
from pybag.encoding.cdr import CdrEncoder
from pybag.mcap.records import ChannelRecord, SchemaRecord
from pybag.schema import SchemaEncoder
from pybag.schema.ros2msg import (
    Array,
    Complex,
    Primitive,
    Ros2MsgSchemaEncoder,
    Schema,
    SchemaConstant,
    SchemaField,
    Sequence,
    String
)
from pybag.types import Message


class MessageSerializer:
    """Serialize dataclass messages using a schema and message encoder."""

    def __init__(
        self,
        schema_encoder: SchemaEncoder,
        message_encoder: type[MessageEncoder],
    ) -> None:
        self._schema_encoder = schema_encoder
        self._message_encoder = message_encoder

    @property
    def schema_encoding(self) -> str:
        return self._schema_encoder.encoding()

    @property
    def message_encoding(self) -> str:
        return self._message_encoder.encoding()

    def _encode_field(
        self,
        encoder: MessageEncoder,
        value: Any,
        schema_field: SchemaField,
        sub_schemas: dict[str, Schema],
    ) -> None:
        if isinstance(schema_field.type, Primitive):
            encoder.encode(schema_field.type.type, value)
            return

        if isinstance(schema_field.type, String):
            encoder.encode(schema_field.type.type, value)
            return

        if isinstance(schema_field.type, Array):
            array_type = schema_field.type
            if isinstance(array_type.type, (Primitive, String)):
                encoder.array(array_type.type.type, value)
                return
            if isinstance(array_type.type, Complex):
                for item in value:
                    self._encode_message(
                        encoder,
                        item,
                        sub_schemas[array_type.type.type],
                        sub_schemas,
                    )
                return
            raise ValueError(f"Unknown array type: {array_type.type}")

        if isinstance(schema_field.type, Sequence):
            sequence_type = schema_field.type
            if isinstance(sequence_type.type, (Primitive, String)):
                encoder.sequence(sequence_type.type.type, value)
                return
            if isinstance(sequence_type.type, Complex):
                encoder.uint32(len(value))
                for item in value:
                    self._encode_message(
                        encoder,
                        item,
                        sub_schemas[sequence_type.type.type],
                        sub_schemas,
                    )
                return
            raise ValueError(f"Unknown sequence type: {sequence_type.type}")

        if isinstance(schema_field.type, Complex):
            complex_type = schema_field.type
            if complex_type.type not in sub_schemas:
                raise ValueError(
                    f"Complex type {complex_type.type} not found in sub_schemas"
                )
            self._encode_message(
                encoder,
                value,
                sub_schemas[complex_type.type],
                sub_schemas,
            )
            return

        raise ValueError(f"Unknown field type: {schema_field.type}")

    def _encode_message(
        self,
        encoder: MessageEncoder,
        message: Message,
        schema: Schema,
        sub_schemas: dict[str, Schema],
    ) -> None:
        for field_name, schema_field in schema.fields.items():
            if isinstance(schema_field, SchemaConstant):
                continue
            if isinstance(schema_field, SchemaField):
                value = getattr(message, field_name)
                self._encode_field(encoder, value, schema_field, sub_schemas)
                continue
            raise ValueError(f"Unknown schema field type: {schema_field}")

    def serialize_message(self, message: Message, *, little_endian: bool = True) -> bytes:
        if not is_dataclass(message):  # pragma: no cover - defensive programming
            raise TypeError("Expected a dataclass instance")
        encoder = self._message_encoder(little_endian=little_endian)
        schema, sub_schemas = self._schema_encoder.parse_schema(message)
        if isinstance(schema, Complex):
            schema = sub_schemas[schema.type]
        self._encode_message(encoder, message, schema, sub_schemas)
        return encoder.save()

    def serialize_schema(self, schema: type[Message]) -> bytes:
        return self._schema_encoder.encode(schema)


class MessageSerializerFactory:
    """Factory for creating message serializers."""

    @staticmethod
    def from_profile(profile: str) -> MessageSerializer | None:
        if profile == "ros2":
            return MessageSerializer(Ros2MsgSchemaEncoder(), CdrEncoder)
        return None

    @staticmethod
    def from_channel(channel: ChannelRecord, schema: SchemaRecord) -> MessageSerializer | None:
        if channel.message_encoding == "cdr" and schema.encoding == "ros2msg":
            return MessageSerializer(Ros2MsgSchemaEncoder(), CdrEncoder)
        return None



"""Utilities for writing MCAP files."""

import logging
import zlib
from pathlib import Path
from typing import Any, Callable, Literal

from pybag import __version__
from pybag.io.raw_writer import BaseWriter, BytesWriter, CrcWriter, FileWriter
from pybag.mcap.record_writer import McapRecordWriter
from pybag.mcap.records import (
    ChannelRecord,
    ChunkIndexRecord,
    ChunkRecord,
    DataEndRecord,
    HeaderRecord,
    MessageIndexRecord,
    MessageRecord,
    RecordType,
    SchemaRecord,
    StatisticsRecord,
    SummaryOffsetRecord
)
from pybag.serialize import MessageSerializerFactory
from pybag.types import Message

logger = logging.getLogger(__name__)


class McapFileWriter:
    """High level writer for producing MCAP files."""

    def __init__(
        self,
        writer: BaseWriter,
        *,
        profile: str = "ros2",
        chunk_size: int | None = None,
        chunk_compression: Literal["lz4", "zstd"] | None = None,
    ) -> None:
        self._writer = CrcWriter(writer)
        self._chunk_count = 0
        self._chunk_size = chunk_size
        self._chunk_compression = chunk_compression or ""
        self._compress_chunk = self._create_chunk_compressor()

        self._next_schema_id = 1  # Schema ID must be non-zero
        self._next_channel_id = 1

        self._topics: dict[str, int] = {}  # topic -> channel_id
        self._schemas: dict[type[Message], int] = {}  # type -> schema_id
        self._schema_records: list[SchemaRecord] = []
        self._channel_records: list[ChannelRecord] = []
        self._sequences: dict[int, int] = {}

        self._message_count = 0
        self._message_start_time: int | None = None
        self._message_end_time: int | None = None
        self._channel_message_counts: dict[int, int] = {}

        self._current_chunk_buffer: BytesWriter = BytesWriter()
        self._current_chunk_start_time: int | None = None
        self._current_chunk_end_time: int | None = None
        self._current_message_index: dict[int, list[tuple[int, int]]] = {}
        self._chunk_indexes: list[ChunkIndexRecord] = []

        # Write the start of the file
        McapRecordWriter.write_magic_bytes(self._writer)
        header = HeaderRecord(profile=profile, library=f"pybag {__version__}")
        McapRecordWriter.write_header(self._writer, header)

        self._profile = profile
        self._message_serializer = MessageSerializerFactory.from_profile(self._profile)
        if self._message_serializer is None:
            raise ValueError(f"Unknown encoding type: {self._profile}")

    def __enter__(self) -> "McapFileWriter":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()

    @classmethod
    def open(
        cls,
        file_path: str | Path,
        *,
        profile: str = "ros2",
        chunk_size: int | None = None,
        chunk_compression: Literal["lz4", "zstd"] | None = "lz4",
    ) -> "McapFileWriter":
        """Create a writer backed by a file on disk.

        Args:
            file_path: The path to the file to write to.
            profile: The profile to use for the MCAP file.
            chunk_size: The size of the chunk to write to in bytes.
            chunk_compression: The compression to use for the chunk.

        Returns:
            A writer backed by a file on disk.
        """

        return cls(
            FileWriter(file_path),
            profile=profile,
            chunk_size=chunk_size,
            chunk_compression=chunk_compression,
        )

    def add_channel(self, topic: str, channel_type: type[Message]) -> int:
        """Add a channel to the MCAP output.

        Args:
            topic: The topic name.
            channel_type: The type of the messages to be written to the channel.

        Returns:
            The channel ID.
        """
        if (channel_id := self._topics.get(topic)) is None:
            # Register the schema if it's not already registered
            if (schema_id := self._schemas.get(channel_type)) is None:
                schema_id = self._next_schema_id
                self._next_schema_id += 1

                # Check that the channel type has a __msg_name__ attribute
                if not isinstance(channel_type, Message):
                    raise ValueError(f"Channel type {channel_type} needs a __msg_name__ attribute")

                schema_record = SchemaRecord(
                    id=schema_id,
                    name=channel_type.__msg_name__,
                    encoding=self._message_serializer.schema_encoding,
                    data=self._message_serializer.serialize_schema(channel_type),
                )

                McapRecordWriter.write_schema(self._writer, schema_record)
                self._schemas[channel_type] = schema_id
                self._schema_records.append(schema_record)

            # Register the channel if it's not already registered
            channel_id = self._next_channel_id
            self._next_channel_id += 1
            channel_record = ChannelRecord(
                id=channel_id,
                schema_id=schema_id,
                topic=topic,
                message_encoding=self._message_serializer.message_encoding,
                metadata={},
            )
            McapRecordWriter.write_channel(self._writer, channel_record)
            self._topics[topic] = channel_id
            self._channel_records.append(channel_record)
            self._sequences[channel_id] = 0
            self._channel_message_counts[channel_id] = 0

        return channel_id

    def write_message(self, topic: str, timestamp: int, message: Message) -> None:
        """Write a message to a topic at a given timestamp.

        Args:
            topic: The topic name.
            timestamp: The timestamp of the message.
            message: The message to write.
        """

        channel_id = self.add_channel(topic, type(message))
        sequence = self._sequences[channel_id]
        self._sequences[channel_id] = sequence + 1

        record = MessageRecord(
            channel_id=channel_id,
            sequence=sequence,
            log_time=timestamp,
            publish_time=timestamp,
            data=self._message_serializer.serialize_message(message),
        )

        if self._chunk_size is not None:
            if self._current_chunk_start_time is None:
                self._current_chunk_start_time = timestamp
            self._current_chunk_end_time = max(self._current_chunk_end_time or timestamp, timestamp)
            offset = self._current_chunk_buffer.size()
            McapRecordWriter.write_message(self._current_chunk_buffer, record)
            self._current_message_index.setdefault(channel_id, []).append((timestamp, offset))
            if self._current_chunk_buffer.size() >= self._chunk_size:
                self._flush_chunk()
        else:
            McapRecordWriter.write_message(self._writer, record)

        # Keep track of statistics
        self._message_count += 1
        self._channel_message_counts[channel_id] += 1
        self._message_start_time = min(self._message_start_time or timestamp, timestamp)
        self._message_end_time = max(self._message_end_time or timestamp, timestamp)

    def _create_chunk_compressor(self) -> Callable[[bytes], bytes]:
        if self._chunk_compression == "lz4":
            import lz4.frame
            return lz4.frame.compress
        elif self._chunk_compression == "zstd":
            import zstandard as zstd
            return zstd.ZstdCompressor().compress
        elif self._chunk_compression == "":
            return lambda x: x
        else:
            raise ValueError(f"Unsupported chunk compression: {self._chunk_compression}")

    def _flush_chunk(self) -> None:
        records = self._current_chunk_buffer.as_bytes()
        chunk = ChunkRecord(
            message_start_time=self._current_chunk_start_time or 0,
            message_end_time=self._current_chunk_end_time or 0,
            uncompressed_size=len(records),
            uncompressed_crc=zlib.crc32(records),
            compression=self._chunk_compression,
            records=self._compress_chunk(records),
        )
        chunk_start_offset = self._writer.tell()
        McapRecordWriter.write_chunk(self._writer, chunk)
        chunk_length = self._writer.tell() - chunk_start_offset

        # Write the message index records for the chunk
        message_index_offsets = {}
        message_index_start_offset = self._writer.tell()
        for cid, records in self._current_message_index.items():
            message_index_offsets[cid] = self._writer.tell()
            message_index_record = MessageIndexRecord(channel_id=cid, records=records)
            McapRecordWriter.write_message_index(self._writer, message_index_record)
        message_index_length = self._writer.tell() - message_index_start_offset

        self._chunk_indexes.append(
            ChunkIndexRecord(
                message_start_time=chunk.message_start_time,
                message_end_time=chunk.message_end_time,
                chunk_start_offset=chunk_start_offset,
                chunk_length=chunk_length,
                message_index_offsets=message_index_offsets,
                message_index_length=message_index_length,
                compression=chunk.compression,
                compressed_size=len(chunk.records),
                uncompressed_size=chunk.uncompressed_size,
            )
        )
        self._chunk_count += 1
        self._current_chunk_buffer.clear()
        self._current_chunk_start_time = None
        self._current_chunk_end_time = None
        self._current_message_index = {}

    def close(self) -> None:
        if self._current_chunk_buffer.size() > 0:
            self._flush_chunk()

        # Data end record
        data_end = DataEndRecord(data_section_crc=self._writer.get_crc())
        McapRecordWriter.write_data_end(self._writer, data_end)

        summary_start = self._writer.tell()
        self._writer.clear_crc()

        # Schema records
        schema_group_start = summary_start
        for record in self._schema_records:
            McapRecordWriter.write_schema(self._writer, record)
        schema_group_length = self._writer.tell() - schema_group_start

        # Channel records
        channel_group_start = self._writer.tell()
        for record in self._channel_records:
            McapRecordWriter.write_channel(self._writer, record)
        channel_group_length = self._writer.tell() - channel_group_start

        # Message index and chunk index records
        chunk_index_group_start = self._writer.tell()
        for record in self._chunk_indexes:
            McapRecordWriter.write_chunk_index(self._writer, record)
        chunk_index_group_length = self._writer.tell() - chunk_index_group_start

        # Statistics record
        statistics_group_start = self._writer.tell()
        stats = StatisticsRecord(
            message_count=self._message_count,
            schema_count=len(self._schema_records),
            channel_count=len(self._channel_records),
            attachment_count=0,
            metadata_count=0,
            chunk_count=self._chunk_count,
            message_start_time=self._message_start_time or 0,
            message_end_time=self._message_end_time or 0,
            channel_message_counts=self._channel_message_counts,
        )
        McapRecordWriter.write_statistics(self._writer, stats)
        statistics_group_length = self._writer.tell() - statistics_group_start

        # Summary offsets
        summary_offset_start = self._writer.tell()
        if schema_group_length > 0:
            McapRecordWriter.write_summary_offset(
                self._writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.SCHEMA,
                    group_start=schema_group_start,
                    group_length=schema_group_length,
                ),
            )
        if channel_group_length > 0:
            McapRecordWriter.write_summary_offset(
                self._writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.CHANNEL,
                    group_start=channel_group_start,
                    group_length=channel_group_length,
                ),
            )
        if chunk_index_group_length > 0:
            McapRecordWriter.write_summary_offset(
                self._writer,
                SummaryOffsetRecord(
                    group_opcode=RecordType.CHUNK_INDEX,
                    group_start=chunk_index_group_start,
                    group_length=chunk_index_group_length,
                ),
            )
        McapRecordWriter.write_summary_offset(
            self._writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.STATISTICS,
                group_start=statistics_group_start,
                group_length=statistics_group_length,
            ),
        )

        # Write footer record manually for CRC calculation
        self._writer.write(McapRecordWriter._encode_record_type(RecordType.FOOTER))
        self._writer.write(McapRecordWriter._encode_uint64(20))
        self._writer.write(McapRecordWriter._encode_uint64(summary_start))
        self._writer.write(McapRecordWriter._encode_uint64(summary_offset_start))
        self._writer.write(McapRecordWriter._encode_uint32(self._writer.get_crc()))

        # Write the magic bytes again
        McapRecordWriter.write_magic_bytes(self._writer)

        # Close the file
        self._writer.close()



import logging
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import Any, Callable

from pybag.deserialize import MessageDeserializerFactory
from pybag.mcap.error import (
    McapUnknownEncodingError,
    McapUnknownSchemaError,
    McapUnknownTopicError
)
from pybag.mcap.record_reader import (
    BaseMcapRecordReader,
    McapRecordReaderFactory
)

# GLOBAL TODOs:
# - TODO: Add tests with mcaps
# - TODO: Improve performance by batching the reads (maybe)
# - TODO: Do something with CRC
# - TODO: Generate summary section of mcap file
logger = logging.getLogger(__name__)


@dataclass(slots=True)
class DecodedMessage():
    channel_id: int
    sequence: int
    log_time: int
    publish_time: int
    data: Any  # TODO: Figure out how to type this


class McapFileReader:
    """Class to read MCAP file"""

    def __init__(self, reader: BaseMcapRecordReader):
        self._reader = reader

        header = self._reader.get_header()
        self._profile = header.profile
        self._message_deserializer = MessageDeserializerFactory.from_profile(self._profile)

    @staticmethod
    def from_file(file_path: Path | str) -> 'McapFileReader':
        reader = McapRecordReaderFactory.from_file(file_path)
        return McapFileReader(reader)

    @staticmethod
    def from_bytes(data: bytes) -> 'McapFileReader':
        reader = McapRecordReaderFactory.from_bytes(data)
        return McapFileReader(reader)

    def get_topics(self) -> list[str]:
        """Get all topics in the MCAP file."""
        return [c.topic for c in self._reader.get_channels().values()] # TODO: Use a set?

    def get_message_count(self, topic: str) -> int:
        """Get the number of messages in a given topic."""
        channel_id = self._reader.get_channel_id(topic)
        if channel_id is None:
            raise McapUnknownTopicError(f'Topic {topic} not found in MCAP file')
        return self._reader.get_statistics().channel_message_counts[channel_id]

    @property
    def start_time(self) -> int:
        """Get the start time of the MCAP file in nanoseconds since epoch."""
        return self._reader.get_statistics().message_start_time

    @property
    def end_time(self) -> int:
        """Get the end time of the MCAP file in nanoseconds since epoch."""
        return self._reader.get_statistics().message_end_time

    # Message Access

    def messages(
        self,
        topic: str,
        start_time: float | None = None,
        end_time: float | None = None,
        filter: Callable[[DecodedMessage], bool] | None = None,
    ) -> Generator[DecodedMessage, None, None]:
        """
        Iterate over messages in the MCAP file.

        Args:
            topic: Topic to filter by.
            start_time: Start time to filter by. If None, start from the beginning of the file.
            end_time: End time to filter by. If None, read to the end of the file.
            filter: Callable used to filter messages. If None, all messages are returned.

        Returns:
            An iterator over DecodedMessage objects.
        """
        channel_id = self._reader.get_channel_id(topic)
        if channel_id is None:
            raise McapUnknownEncodingError(f'Topic {topic} not found in MCAP file')

        channel_record = self._reader.get_channel(channel_id)
        if (message_schema := self._reader.get_channel_schema(channel_id)) is None:
            raise McapUnknownSchemaError(f'Unknown schema for channel {channel_id}')

        if (message_deserializer := self._message_deserializer) is None:
            message_deserializer = MessageDeserializerFactory.from_channel(
                channel_record, message_schema
            )
        if message_deserializer is None:
            raise McapUnknownEncodingError(f'Unknown encoding type: {self._profile}')

        deserialize = message_deserializer.deserialize_message
        for message in self._reader.get_messages(channel_id, start_time, end_time):
            decoded = DecodedMessage(
                message.channel_id,
                message.sequence,
                message.log_time,
                message.publish_time,
                deserialize(message, message_schema),
            )
            if filter is None or filter(decoded):
                yield decoded

    def close(self) -> None:
        """Close the MCAP reader and release all resources."""
        self._reader.close()

    def __enter__(self) -> "McapFileReader":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None
    ) -> None:
        self.close()



import argparse


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="pybag",
        description=(
            "Command line interface for pybag. Provides utilities to work with "
            "MCAP and bag files."
        ),
    )
    subparsers = parser.add_subparsers(dest="command")

    # In the future subcommands will be added here.

    parser.set_defaults(func=lambda args: parser.print_help())
    return parser


def main(argv: list[str] | None = None) -> None:
    parser = build_parser()
    args = parser.parse_args(argv)
    args.func(args)


if __name__ == "__main__":
    main()



"""Compile a schema into an efficient message decoder."""
from __future__ import annotations

import re
import struct
from types import SimpleNamespace
from typing import Callable

from pybag.encoding import MessageDecoder
from pybag.schema import (
    Array,
    Complex,
    Primitive,
    Schema,
    SchemaConstant,
    SchemaField,
    Sequence,
    String
)

# Map primitive ROS2 types to struct format characters
_STRUCT_FORMAT = {
    "bool": "?",
    "int8": "b",
    "uint8": "B",
    "int16": "h",
    "uint16": "H",
    "int32": "i",
    "uint32": "I",
    "int64": "q",
    "uint64": "Q",
    "float32": "f",
    "float64": "d",
}
_STRUCT_SIZE = {k: struct.calcsize(v) for k, v in _STRUCT_FORMAT.items()}
_TAB = '    '


def _sanitize(name: str) -> str:
    return re.sub(r"[^0-9a-zA-Z_]", "_", name)


def compile_schema(schema: Schema, sub_schemas: dict[str, Schema]) -> Callable[[MessageDecoder], type]:
    """Compile ``schema`` into a decoder function.

    The returned function accepts a :class:`MessageDecoder` instance and returns a
    dynamically constructed type with the decoded fields as class attributes.
    """

    function_defs: list[str] = []
    compiled: dict[str, str] = {}

    def build(current: Schema) -> str:
        func_name = f"decode_{_sanitize(current.name)}"
        if func_name in compiled:
            return func_name

        compiled[func_name] = func_name
        lines: list[str] = [
            f"def {func_name}(decoder):",
            f"{_TAB}fmt_prefix = '<' if decoder._is_little_endian else '>'",
            f"{_TAB}_data = decoder._data",
            f"{_TAB}obj = SimpleNamespace()",
        ]
        field_names: list[str] = []
        run_type: str | None = None
        run_fields: list[str] = []

        def flush() -> None:
            nonlocal run_type, run_fields
            if not run_fields:
                return

            count = len(run_fields)
            size = _STRUCT_SIZE[run_type]  # type: ignore[index]
            fmt = _STRUCT_FORMAT[run_type] * count  # type: ignore[index]
            lines.append(f"{_TAB}_data.align({size})")

            if count > 1:
                names = ", ".join(run_fields)
                lines.append(f"{_TAB}{names} = struct.unpack(fmt_prefix + '{fmt}', _data.read({size * count}))")
                for field in run_fields:
                    lines.append(f"{_TAB}obj.{field} = {field}")
            else:
                field = run_fields[0]
                lines.append(f"{_TAB}obj.{field} = struct.unpack(fmt_prefix + '{fmt}', _data.read({size * count}))[0]")

            run_fields = []
            run_type = None

        for field_name, entry in current.fields.items():
            field_names.append(field_name)
            if isinstance(entry, SchemaConstant):
                flush()
                lines.append(f"{_TAB}obj.{field_name} = {repr(entry.value)}")
                continue

            if not isinstance(entry, SchemaField):
                flush()
                lines.append(f"{_TAB}obj.{field_name} = None")
                continue

            field_type = entry.type

            if isinstance(field_type, Primitive) and field_type.type in _STRUCT_FORMAT:
                if run_type == field_type.type:
                    run_fields.append(field_name)
                else:
                    flush()
                    run_type = field_type.type
                    run_fields = [field_name]
                continue

            flush()

            if isinstance(field_type, Primitive):
                lines.append(f"{_TAB}obj.{field_name} = decoder.{field_type.type}()")

            elif isinstance(field_type, String):
                lines.append(f"{_TAB}obj.{field_name} = decoder.{field_type.type}()")

            elif isinstance(field_type, Array):
                elem = field_type.type
                if isinstance(elem, Primitive) and elem.type in _STRUCT_FORMAT:
                    size = _STRUCT_SIZE[elem.type]
                    fmt = _STRUCT_FORMAT[elem.type] * field_type.length
                    lines.append(f"{_TAB}_data.align({size})")
                    lines.append(
                        f"{_TAB}obj.{field_name} = list(struct.unpack(fmt_prefix + '{fmt}', _data.read({size * field_type.length})))"
                    )
                elif isinstance(elem, Complex):
                    sub_schema = sub_schemas[elem.type]
                    sub_func = build(sub_schema)
                    lines.append(
                        f"{_TAB}obj.{field_name} = [{sub_func}(decoder) for _ in range({field_type.length})]"
                    )
                elif isinstance(elem, String):
                    elem_name = elem.type
                    lines.append(
                        f"{_TAB}obj.{field_name} = [decoder.{elem_name}() for _ in range({field_type.length})]"
                    )
                else:
                    elem_name = getattr(elem, "type", "unknown")
                    lines.append(
                        f"{_TAB}obj.{field_name} = decoder.array('{elem_name}', {field_type.length})"
                    )

            elif isinstance(field_type, Sequence):
                elem = field_type.type
                if isinstance(elem, Primitive) and elem.type in _STRUCT_FORMAT:
                    size = _STRUCT_SIZE[elem.type]
                    char = _STRUCT_FORMAT[elem.type]
                    lines.append(f"{_TAB}_len = decoder.uint32()")
                    lines.append(f"{_TAB}_data.align({size})")
                    lines.append(
                        f"{_TAB}obj.{field_name} = list(struct.unpack(fmt_prefix + '{char}' * _len, _data.read({size} * _len)))"
                    )
                elif isinstance(elem, Complex):
                    sub_schema = sub_schemas[elem.type]
                    sub_func = build(sub_schema)
                    lines.append(f"{_TAB}length = decoder.uint32()")
                    lines.append(
                        f"{_TAB}obj.{field_name} = [{sub_func}(decoder) for _ in range(length)]"
                    )
                elif isinstance(elem, String):
                    lines.append(f"{_TAB}length = decoder.uint32()")
                    elem_name = elem.type
                    lines.append(
                        f"{_TAB}obj.{field_name} = [decoder.{elem_name}() for _ in range(length)]"
                    )
                else:
                    elem_name = getattr(elem, "type", "unknown")
                    lines.append(f"{_TAB}obj.{field_name} = decoder.sequence('{elem_name}')")

            elif isinstance(field_type, Complex):
                sub_schema = sub_schemas[field_type.type]
                sub_func = build(sub_schema)
                lines.append(f"{_TAB}obj.{field_name} = {sub_func}(decoder)")

            else:
                lines.append(f"{_TAB}obj.{field_name} = None")

        flush()
        lines.append(f"{_TAB}return obj")
        function_defs.append("\n".join(lines))
        return func_name

    build(schema)
    code = "import struct\nfrom types import SimpleNamespace\n" + "\n\n".join(function_defs)
    namespace: dict[str, object] = {"struct": struct, "SimpleNamespace": SimpleNamespace}
    exec(code, namespace)
    return namespace[f"decode_{_sanitize(schema.name)}"]  # type: ignore[index]

__all__ = ["compile_schema"]
